//
//  OmnisJSAppDelegate.m
//  OmnisJS
//
//  Created by Bob Mitchell on 31/05/2011.
//  Copyright 2011 TigerLogic UK Ltd. All rights reserved.
//

#import "OmnisJSAppDelegate.h"
#import "OmnisJSViewController.h"
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>

@implementation OmnisJSAppDelegate

@synthesize window;
@synthesize viewController;


#pragma mark -
#pragma mark Application lifecycle

- (BOOL)application:(UIApplication *)application didFinishLaunchingWithOptions:(NSDictionary *)launchOptions 
{        
    // Override point for customization after application launch.
		mTestFormThread = nil;
		mStopTestFormThread = NO;
		// Read preferences
		if ([self designModeEnabled])
		{
			mShowTitle = YES;
			[self applicationWillEnterForeground:application];
		}
		else
		{
			// Read configuration from the bundle, if it is present
			NSString *filePath = [[NSBundle mainBundle] pathForResource:@"config" ofType:@"xml"];
			NSURL *fileURL = [[NSURL alloc] initFileURLWithPath:filePath isDirectory:NO];
			NSXMLParser *parser = [[NSXMLParser alloc] initWithContentsOfURL:fileURL];
			[parser setDelegate:self];
			mInURL = NO;
			mURL = [[NSMutableString alloc] init];
			[parser parse];
			[parser release];
			[fileURL release];			
		}
				
		[[UIApplication sharedApplication] setStatusBarHidden:!mShowTitle];

    // Add the view controller's view to the window and display.
    [self.window addSubview:viewController.view];
    [self.window makeKeyAndVisible];

    return YES;
}


- (void)applicationWillResignActive:(UIApplication *)application 
{
    /*
     Sent when the application is about to move from active to inactive state. This can occur for certain types of temporary interruptions (such as an incoming phone call or SMS message) or when the user quits the application and it begins the transition to the background state.
     Use this method to pause ongoing tasks, disable timers, and throttle down OpenGL ES frame rates. Games should use this method to pause the game.
     */
}


- (void)applicationDidEnterBackground:(UIApplication *)application 
{
    /*
     Use this method to release shared resources, save user data, invalidate timers, and store enough application state information to restore your application to its current state in case it is terminated later. 
     If your application supports background execution, called instead of applicationWillTerminate: when the user quits.
     */
		@synchronized(self)
		{
			if (mTestFormThread)
			{
				mStopTestFormThread = true;
				mTestFormThread = nil;
			}
		}
}


- (void)applicationWillEnterForeground:(UIApplication *)application 
{
    /*
     Called as part of  transition from the background to the inactive state: here you can undo many of the changes made on entering the background.
     */
		@synchronized(self)
		{
			if ([self designModeEnabled])
			{
				NSUserDefaults *ud = [NSUserDefaults standardUserDefaults];
				mDesignModeServer = [ud stringForKey:@"designserver"];
				[self startTestFormThread];
			}
		}
}


- (void)applicationDidBecomeActive:(UIApplication *)application 
{
    /*
     Restart any tasks that were paused (or not yet started) while the application was inactive. If the application was previously in the background, optionally refresh the user interface.
     */
}


- (void)applicationWillTerminate:(UIApplication *)application 
{
    /*
     Called when the application is about to terminate.
     See also applicationDidEnterBackground:.
     */
}


#pragma mark -
#pragma mark Memory management

- (void)applicationDidReceiveMemoryWarning:(UIApplication *)application 
{
    /*
     Free up as much memory as possible by purging cached data objects that can be recreated (or reloaded from disk) later.
     */
}


- (void)dealloc 
{
    [viewController release];
    [window release];
    [super dealloc];
}

// NSXMLParser delegate methods
- (void)parser:(NSXMLParser *)parser didStartElement:(NSString *)elementName namespaceURI:(NSString *)namespaceURI qualifiedName:(NSString *)qualifiedName attributes:(NSDictionary *)attributeDict
{
	if ([elementName caseInsensitiveCompare:@"URL"] == NSOrderedSame)
	{
		mInURL = YES;
		mShowTitle = ([[attributeDict objectForKey:@"title"] integerValue] != 0);
	}
}

- (void)parser:(NSXMLParser *)parser didEndElement:(NSString *)elementName namespaceURI:(NSString *)namespaceURI qualifiedName:(NSString *)qName
{
	if ([elementName compare:@"URL"])
	{
		mInURL = NO;
	}
}

- (void)parser:(NSXMLParser *)parser foundCharacters:(NSString *)string
{
	[mURL appendString:string];
}

// Miscellaneous methods
- (void)loadURL 
{
	[[UIApplication sharedApplication] setStatusBarHidden:!mShowTitle];
	[viewController setWebViewFrame:[[UIScreen mainScreen] applicationFrame]];
	if ([mURL length] > 0)
	{
		NSString *url = [mURL stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceAndNewlineCharacterSet]];
		if ([url length] > 0)
			[viewController loadURL:url];
	}
	[mURL release];
	mURL = nil;
}

// Test form thread
-(BOOL)designModeEnabled
{
	NSUserDefaults *ud = [NSUserDefaults standardUserDefaults];
	return [ud boolForKey:@"designmodeon"];
}

-(void)startTestFormThread
{
	@synchronized(self)
	{
		mTestFormThread = [[NSThread alloc] initWithTarget:self selector:@selector(threadEntry) object:nil];
		[mTestFormThread start];
	}
}

-(void)threadEntry
{
	// Get the IP address and port of the design mode server
	sockaddr_in remoteAddress;
	memset(&remoteAddress, 0, sizeof(remoteAddress));
	remoteAddress.sin_family = PF_INET;
	
	const char *serverUtf8 = [mDesignModeServer UTF8String];
	if (!serverUtf8)
		return;	// Should never get to here
	char *server = strdup(serverUtf8);
	char *portPtr = strchr(server, ':');
	if (portPtr)
	{
		*portPtr++ = 0;
	}
	free(server);
	
	remoteAddress.sin_addr.s_addr = inet_addr(server);
	remoteAddress.sin_port = htons((unsigned short) (portPtr ? atoi(portPtr) : 5000));

	BOOL stopThread = NO;
	int serverSocket = -1;
	BOOL first = YES;
	BOOL connected = NO;
	BOOL needAbort = NO;
	BOOL messageSent = NO;
	char receiveBuffer[4096];
	int bytesReceived = 0;
	int messageBytesNeeded = 0;
	do
	{
		@synchronized(self)
		{
			stopThread = mStopTestFormThread;
			mStopTestFormThread = NO;
		}
		if (stopThread)
		{
			break;
		}
		if (first)
		{
			first = NO;
		}
		else
		{
			// Sleep for 250 milliseconds
			[NSThread sleepForTimeInterval:0.25];
		}
		// Abort if required
		if (needAbort)
		{
			[self abortSocket:serverSocket];
			serverSocket = -1;
			needAbort = NO;
			connected = NO;
			messageSent = NO;
			bytesReceived = messageBytesNeeded = 0;
		}
		// Create a socket if required
		if (-1 == serverSocket)
		{
			serverSocket = socket(PF_INET, SOCK_STREAM, 0);
			if (serverSocket == -1)
				continue;
			// Got the socket
			if (![self setNonBlocking:NO socket:serverSocket])
			{
				needAbort = YES;
				continue;
			}
		}
		if (!connected)
		{
			// Got a socket: connect to the server
			if (-1 == ::connect(serverSocket, (const sockaddr *) &remoteAddress, sizeof(struct sockaddr)))
			{
				// Try again after a delay
				continue;
			}
			connected = YES;
			if (![self setNonBlocking:YES socket:serverSocket])
			{
				needAbort = YES;
				continue;
			}
		}
		if (!messageSent)
		{
			// Build initial message to set up test form mode with server
			char connectMessage[23];
			memset(connectMessage, 0, sizeof(connectMessage));
			connectMessage[3] = 23;	// Overall message length
			memcpy(&connectMessage[4], "OMNIS_ORFC", 10);
			connectMessage[15] = 1;	// Version
			connectMessage[17] = 40;// JavaScript test form connect message
		
			int bytesToSend = sizeof(connectMessage);
			char *buffer = connectMessage;
			while (bytesToSend > 0)
			{
				int thisSend = bytesToSend;
				int numSent = send(serverSocket, buffer, thisSend, 0);
				if (-1 == numSent)
				{
					if (EAGAIN == errno)
					{
						// Call would block
						[NSThread sleepForTimeInterval:0.001];
						continue;
					}
					needAbort = YES;
					break;
				}
				buffer += numSent;
				bytesToSend -= numSent;
			}
			if (bytesToSend > 0)
				continue;
			messageSent = YES;
		}
		// Next, receive the response
		if (bytesReceived < 4)
		{
			int thisReceive = recv(serverSocket, &receiveBuffer[bytesReceived], 4 - bytesReceived, 0);
			if (-1 == thisReceive)
			{
				if (EAGAIN != errno)
				{
					needAbort = YES;
					continue;
				}
				// Call would block
				continue;
			}
			else
			{
				bytesReceived += thisReceive;
			}
			if (bytesReceived >= 4)
			{
				messageBytesNeeded = (((int) receiveBuffer[0]) << 24) + (((int) receiveBuffer[1]) << 16) + (((int) receiveBuffer[2]) << 8) + ((int) receiveBuffer[3]);
				if (messageBytesNeeded < 20)	// 18 for header, 2 for title indicator and file name
				{
					needAbort = YES;
					continue;
				}
			}
		}
		// Receive rest of message
		if (bytesReceived < messageBytesNeeded)
		{
			int thisReceive = recv(serverSocket, &receiveBuffer[bytesReceived], messageBytesNeeded - bytesReceived, 0);
			if (-1 == thisReceive)
			{
				if (EAGAIN != errno)
				{
					needAbort = YES;
					continue;
				}
				// Call would block
				continue;
			}
			else
			{
				bytesReceived += thisReceive;
			}
		}
		// Process message
		if (bytesReceived == messageBytesNeeded)
		{
			// First byte is the title indicator
			mShowTitle = (receiveBuffer[18] == '1');
			// Load URL
			receiveBuffer[messageBytesNeeded] = 0; // Null terminator
			NSString *fileName = [[NSString alloc] initWithUTF8String:&receiveBuffer[19]];
			mURL = [NSMutableString stringWithFormat:@"http://%@/jschtml/%@", mDesignModeServer, fileName];			
			[self performSelectorOnMainThread:@selector(loadURL) withObject:nil waitUntilDone:NO];

			// Ready for next test form
			bytesReceived = 0;
		}
	} while (1);
}

-(void)abortSocket:(int)pSocket 
{
	struct linger lingerInfo;
	lingerInfo.l_onoff = 1;
	lingerInfo.l_linger = 0;
	setsockopt(pSocket, SOL_SOCKET, SO_LINGER, (char *) &lingerInfo, sizeof(lingerInfo));
	close(pSocket);	
}

-(BOOL)setNonBlocking:(BOOL)pNonBlocking socket:(int)socket
{
	// Make the socket non-blocking or blocking
	BOOL done = NO;
	int flags = fcntl(socket, F_GETFL, 0);
	if (-1 != flags)
	{
		if (pNonBlocking) flags |= O_NONBLOCK;
		else flags &= ~O_NONBLOCK;
		if (-1 != fcntl(socket, F_SETFL, flags))
		{
			done = YES;
		}
	}
	return done;
}

@end
